# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 - 2022 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for generic and customizable requests."""

from invenio_i18n import lazy_gettext as _
from invenio_users_resources.entity_resolvers import GroupResolver, UserResolver

from invenio_requests.services.requests import facets

from .customizations import CommentEventType, LogEventType
from .services.permissions import PermissionPolicy

REQUESTS_PERMISSION_POLICY = PermissionPolicy
"""Override the default requests/comments permission policy."""

REQUESTS_REGISTERED_TYPES = []
"""Configuration for registered Request Types."""

REQUESTS_REGISTERED_EVENT_TYPES = [
    LogEventType(),
    CommentEventType(),
]
"""Configuration for registered Request Event Types."""

REQUESTS_ENTITY_RESOLVERS = [UserResolver(), GroupResolver()]
"""Registered resolvers for resolving/creating references in request metadata."""

REQUESTS_ROUTES = {
    "details": "/requests/<pid_value>",
}
"""Invenio requests ui endpoints."""

REQUESTS_SEARCH = {
    "facets": ["type", "status"],
    "sort": ["bestmatch", "newest", "oldest"],
}
"""Requests search default configuration."""

REQUESTS_SORT_OPTIONS = {
    "bestmatch": dict(
        title=_("Best match"),
        fields=["_score"],  # search defaults to desc on `_score` field
    ),
    "newest": dict(
        title=_("Newest"),
        fields=["-created"],
    ),
    "oldest": dict(
        title=_("Oldest"),
        fields=["created"],
    ),
}
"""Definitions of available request sort options."""

REQUESTS_FACETS = {
    "type": {
        "facet": facets.type,
        "ui": {
            "field": "type",
        },
    },
    "status": {
        "facet": facets.status,
        "ui": {
            "field": "status",
        },
    },
}
"""Invenio requests facets."""

REQUESTS_TIMELINE_PAGE_SIZE = 15
"""Amount of items per page on the request details timeline"""


REQUESTS_MODERATION_ROLE = "administration-moderation"
"""ID of the Role used for moderation."""


#
# User moderation administration
#
REQUESTS_USER_MODERATION_SEARCH = {
    "facets": ["status", "is_open"],
    "sort": ["bestmatch", "newest", "oldest"],
}
"""Community requests search configuration (i.e list of community requests)"""

REQUESTS_USER_MODERATION_SORT_OPTIONS = {
    "bestmatch": dict(
        title=_("Best match"),
        fields=["_score"],  # ES defaults to desc on `_score` field
    ),
    "newest": dict(
        title=_("Newest"),
        fields=["-created"],
    ),
    "oldest": dict(
        title=_("Oldest"),
        fields=["created"],
    ),
}
"""Definitions of available record sort options."""

REQUESTS_USER_MODERATION_FACETS = {
    "status": {
        "facet": facets.status,
        "ui": {
            "field": "status",
        },
    },
    "is_open": {"facet": facets.is_open, "ui": {"field": "is_open"}},
}
"""Available facets defined for this module."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Base classes for requests in Invenio."""

import uuid

from invenio_db import db
from invenio_records.models import RecordMetadataBase
from sqlalchemy import func
from sqlalchemy.dialects import mysql
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.sql import text
from sqlalchemy.types import String
from sqlalchemy_utils import UUIDType


class RequestMetadata(db.Model, RecordMetadataBase):
    """Base class for requests of any kind in Invenio."""

    __tablename__ = "request_metadata"

    id = db.Column(UUIDType, primary_key=True, default=uuid.uuid4)

    number = db.Column(String(50), unique=True, index=True, nullable=True)

    expires_at = db.Column(
        db.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
        default=None,
        nullable=True,
    )


class RequestEventModel(db.Model, RecordMetadataBase):
    """Request Events model."""

    __tablename__ = "request_events"

    type = db.Column(db.String(1), nullable=False)
    request_id = db.Column(
        UUIDType, db.ForeignKey(RequestMetadata.id, ondelete="CASCADE")
    )
    request = db.relationship(RequestMetadata)


class SequenceMixin:
    """Integer sequence generator.

    The purpose of this model is to generate a sequence of integers using the
    underlying database's auto increment features in a transaction friendly
    manner.

    The database table have a single column declared as auto-incrementing. To
    obtain a new value, we insert a row in the table and let the database
    assign a value to it.

    If two transactions try to insert the same value one will fail/retry.

    Usage:

    .. code-block:: python

        class MySequence(db.Model, SequenceMixin):
            __table_name__ = "mycounter"
    """

    @declared_attr
    def value(cls):
        """The counter."""
        return db.Column(
            db.BigInteger().with_variant(db.Integer, "sqlite"),
            primary_key=True,
            autoincrement=True,
        )

    @classmethod
    def next(cls):
        """Return next available integer."""
        try:
            with db.session.begin_nested():
                obj = cls()
                db.session.add(obj)
        except IntegrityError:  # pragma: no cover
            with db.session.begin_nested():
                # Someone has likely modified the table without using the
                # models API. Let's fix the problem.
                cls._set_sequence(cls.max())
                obj = cls()
                db.session.add(obj)
        return obj.value

    @classmethod
    def max(cls):
        """Get max record identifier."""
        max_value = db.session.query(func.max(cls.value)).scalar()
        return max_value if max_value else 0

    @classmethod
    def _set_sequence(cls, val):
        """Internal function to reset sequence to specific value.

        Note: this function is for PostgreSQL compatibility.

        :param val: The value to be set.
        """
        if db.engine.dialect.name == "postgresql":  # pragma: no cover
            db.session.execute(
                text(
                    "SELECT setval(pg_get_serial_sequence(:table, 'value'), :newval)"
                ).bindparams(newval=val, table=cls.__tablename__),
            )

    @classmethod
    def insert(cls, val):
        """Insert a record identifier.

        :param val: The value to insert.
        """
        with db.session.begin_nested():
            obj = cls(value=val)
            db.session.add(obj)
            cls._set_sequence(cls.max())


class RequestNumber(db.Model, SequenceMixin):
    """Request number sequence model."""

    __tablename__ = "request_number_seq"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 CERN.
# Copyright (C) 2021 Northwestern University.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Request permissions."""

import operator
from functools import reduce
from itertools import chain

from invenio_records_permissions.generators import Generator
from invenio_records_resources.references import EntityGrant
from invenio_search.engine import dsl

from invenio_requests.proxies import current_requests


class Status(Generator):
    """Generator to validate needs only for a given request status."""

    def __init__(self, statuses, generators):
        """Initialize need entity generator."""
        self._statuses = statuses
        self._generators = generators or []

    def needs(self, request=None, **kwargs):
        """Needs if status is in one of the provided ones."""
        if request.status in self._statuses:
            needs = [g.needs(request=request, **kwargs) for g in self._generators]
            return set(chain.from_iterable(needs))
        return []

    def query_filter(self, **kwargs):
        """Query filters for the current identity."""
        queries = [g.query_filter(**kwargs) for g in self._generators]
        queries = [q for q in queries if q]
        queries = reduce(operator.or_, queries) if queries else None

        if queries:
            return dsl.Q("terms", **{"status": self._statuses}) & queries
        return None


class EntityNeedsGenerator(Generator):
    """Allows the creator of the request."""

    entity_field = None
    grants_field = "grants"

    def __init__(self):
        """Initialize need entity generator."""
        assert self.entity_field is not None, "Subclass must define entity_field."

    def needs(self, request=None, **kwargs):
        """Needs for the given entity reference."""
        entity = getattr(request, self.entity_field)
        return request.type.entity_needs(entity)

    def query_filter(self, identity=None, **kwargs):
        """Query filters for the current identity."""
        grants = []
        for need in identity.provides:
            grants.append(EntityGrant(self.entity_field, need).token)
        if grants:
            return dsl.Q("terms", **{self.grants_field: grants})
        return None


class Creator(EntityNeedsGenerator):
    """Allows the creator of the request."""

    entity_field = "created_by"


class Receiver(EntityNeedsGenerator):
    """Allows the receiver of the request."""

    entity_field = "receiver"


class Topic(EntityNeedsGenerator):
    """Allows access control based on the request's topic.

    This generator grants permissions to users/entities associated with the topic
    of a request. The specific needs depend on whether the request type supports
    `resolve_topic_needs`.
    """

    entity_field = "topic"

    def needs(self, request=None, **kwargs):
        """Determine the access needs for the given request topic.

        If the request type defines `resolve_topic_needs`, it means the topic's
        permissions should be resolved dynamically based on the entity associated
        with it (e.g., a record, community, or another object).

        Otherwise, the topic is not involved in access control, and no additional
        permissions are granted.
        """
        entity = getattr(request, self.entity_field)

        if getattr(request.type, "resolve_topic_needs"):
            return request.type.entity_needs(entity)

        return []

    def query_filter(self, identity=None, **kwargs):
        """Construct a query filter to include only requests where the topic grants access.

        This filter ensures that only request types that support `resolve_topic_needs`
        contribute to permissions. Request types that do not define this attribute
        are explicitly excluded.

        Excluding request types without `resolve_topic_needs` prevents them from
        granting unintended access based on their topic, as they may not have
        the necessary logic to determine entity-specific permissions.
        """
        # Exclude request types that do NOT define `resolve_topic_needs`
        excluded_request_types = [
            ~dsl.Q("term", **{"type": _type.type_id})
            for _type in current_requests.request_type_registry
            if not getattr(_type, "resolve_topic_needs")
        ]

        # Generate grant tokens based on the user's identity
        grant_tokens = [
            EntityGrant(self.entity_field, need).token for need in identity.provides
        ]

        # If the user has no grant tokens, there is no need to proceed
        if not grant_tokens:
            return None

        # Build the final query
        query = dsl.Q(
            "bool",
            must=[
                dsl.Q("terms", **{self.grants_field: grant_tokens}),
                dsl.Q("bool", must=excluded_request_types),
            ],
        )

        return query


class Commenter(Generator):
    """The user who created a specific comment."""

    def needs(self, event=None, request=None, **kwargs):
        """Enabling Needs."""
        if event.created_by is not None:
            return event.created_by.get_needs()
        return []

    def query_filter(self, identity=None, **kwargs):
        """Filters for current identity as creator."""
        raise RuntimeError("The generator cannot be used for searching.")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 TU Wien.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Status systemfield for request records."""

from invenio_records.systemfields import SystemField


class RequestStatusField(SystemField):
    """Systemfield for managing the request status."""

    def __set__(self, record, value):
        """Set the request's status."""
        assert record is not None

        if value not in record.type.available_statuses:
            raise ValueError(f"Unknown status: '{value}'")

        self.set_dictkey(record, value)

    def __get__(self, record, owner=None):
        """Get the request's status."""
        if record is None:
            # access by class
            return self

        return self.get_dictkey(record)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio-Requests is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Event type systemfield for events records."""

import inspect

from invenio_records.systemfields import ModelField

from ...customizations import EventType
from ...proxies import current_event_type_registry


class EventTypeField(ModelField):
    """Systemfield for managing the request event type."""

    @staticmethod
    def get_instance(value):
        """Ensure that always an instance is returned."""
        if inspect.isclass(value):
            # if a class was passed rather than an instance, try to instantiate it
            value = value()

        if isinstance(value, str):
            value = current_event_type_registry.lookup(value)

        if not isinstance(value, EventType):
            raise TypeError(f"Expected 'EventType' but got: '{type(value)}'")
        return value

    def _set(self, model, value):
        """Initialize model value."""
        value = EventTypeField.get_instance(value)
        super()._set(model, value.type_id)

    def set_obj(self, instance, obj):
        """."""
        self.set_dictkey(instance, obj.type_id)
        self._set_cache(instance, obj)
        self._set(instance.model, obj)

    def __set__(self, record, value):
        """."""
        assert record is not None
        value = EventTypeField.get_instance(value)
        self.set_obj(record, value)

    def obj(self, instance):
        """Get the request type."""
        obj = self._get_cache(instance)
        if obj is not None:
            return obj

        # check model for the type_id
        type_id = super().__get__(instance)
        if not type_id:
            # check in passed instance for the type_id
            type_id = self.get_dictkey(instance)
        obj = EventTypeField.get_instance(type_id)
        self._set_cache(instance, obj)

        return obj

    def __get__(self, record, owner=None):
        """Get the request type."""
        # Class access
        if record is None:
            return self

        return self.obj(record)

    #
    # Record extension
    #
    def pre_init(self, record, data, model=None, **kwargs):
        """Ensure type is always in the registry."""
        _type = kwargs.get("type")
        if _type is None:
            # retrieve type from instantiated model
            _type = model.type
        value = EventTypeField.get_instance(_type).type_id
        try:
            # validate type
            current_event_type_registry.lookup(value)
        except KeyError:
            raise TypeError(f"Event type {value} is not registered.")
